import { defineQuery, hasComponent } from 'bitecs'
import { query as q } from './query-parser/index.js'
import { parseAll } from './parser.js'
import { curry, of, is, map, path, pipe, always, assocPath, prop, identity } from './fn.js'
import { Relationship } from './query-parser/types.js'

const nodeComponent = component => node => {
  const label = node.label?.value
  const name = node.name?.value
  const type = label && component[node.label.value] || identity

  return { name, label, type }
}

const notEntity = ({ label }) => label != null && label !== 'Entity'

const prepareQuery = (query, component) => {
  const { match, returnValues } = query
  const relationship = is(Relationship, match)

  if (relationship) {
    const { from: f, to: t, edge: e } = match
    const [from, to, edge] = map(q => prepareQuery({ match: q, returnValues }, component), [f, t, e])

    return { from, to, edge, relationship, schema: query }
  } else {
    const components = match.components.map(nodeComponent(component))
    const types = components.filter(notEntity).map(prop('type'))
    return { components, relationship, query: defineQuery(types), schema: query }
  }
}

const hasComponents = (world, types, entity) => (
  types.every(type => hasComponent(world, type, entity))
)

const matches = (world, types, set = new Set()) => entity => {
  if (set.has(entity)) {
    return true
  } else if (hasComponents(world, types, entity)) {
    set.add(entity)
    return true
  } else {
    return false
  }
}

const queryRelationship = (query, world) => {
  const [from, edge, to] = ['from', 'edge', 'to'].map(name => {
    const node = query[name]
    return node.relationship ? node.from : node
  })

  const matchesFrom = matches(world, from.components.map(prop('type')))
  const matchesTo = matches(world, to.components.map(prop('type')))

  const Edge = edge.components[0].type
  // FIXME: filter these to make sure the from components match the previous edge's to components
  const edges = edge.query(world)
  console.log('59', edges)
  const result = []
  for (let i = 0; i < edges.length; i++) {
    const eid = edges[i]

    if (matchesFrom(Edge.from[eid])) {
      console.log('65', eid, from.components.map(({ name, label }) => ({ name, label })))
      if (matchesTo(Edge.to[eid])) {
        console.log('67', eid, to.components.map(({ name, label }) => ({ name, label })))
      }
    }
    if (matchesFrom(Edge.from[eid]) && matchesTo(Edge.to[eid])) {
      result.push(eid)
    }
  }

  return result
}

const assembleQuery = (query, entities) => map(entity => ([{ entity, query }]), entities)

const executeQuery = curry((query, world) => {
  if (!query.relationship) {
    return assembleQuery(query, query.query(world))
  }

  const edges = queryRelationship(query, world)
  console.log('80', query.from.components, edges)

  if (query.to.relationship) {
    return edges.reduce((acc, eid) => {
      const next = assocPath(
        'from.query',
        always(of(query.edge.components[0].type.to[eid])),
        query.to
      )

      const a = executeQuery(next, world)
        .map(child => [eid, ...child])
        .map(([l, r]) => [{ entity: l, query }, r])
      //.concat(acc)
      console.log(a)
      return a.concat(acc)
    }, [])
  } else {
    return assembleQuery(query, edges)
  }
})

const isTag = type => (
  type && Object.getOwnPropertySymbols(type).find(
    (s) => s.description === 'tagStore'
  ) || false
)

const getValue = (type, entity) => {
  if (isTag(type)) {
    return {}
  } else if (typeof type === 'function') {
    return type(entity)
  } else {
    return Object.fromEntries(
      Object.entries(type).map(([k, v]) => [k, v[entity]])
    )
  }
}

const resolveNode = curry((node, obj) => {
  const returns = path('query.schema.returnValues', node)
  const { entity, query: { components } } = node
  const values = components.filter(({ name }) => returns.includes(name))
    .map(({ name, type }) => ({ [name]: getValue(type, entity) }))
  return Object.assign(obj, ...values)
})

const resolveRelationship = edges => {
  const { entity, query } = edges
  const Edge = query.edge.components[0].type

  const to = query.to?.from ?? query.to
  return pipe(
    resolveNode({ entity: Edge.from[entity], query: query.from }),
    resolveNode({ entity, query: query.edge }),
    resolveNode({ entity: Edge.to[entity], query: to }),
  )({})
}

const resolveReturns = query => {
  return query.map(edges => {
    const returns = edges.map(edge => {
      if (edge.query.relationship) {
        return resolveRelationship(edge)
      } else {
        return resolveNode(edge, {})
      }
    })

    return Object.assign({}, ...returns)
  })
}

export const query = curry((input, { world, component }) => {
  return parseAll(q, input).map(({ use, ...rest }) => {
    const useWorld = world[use?.value ?? 'default']
    const preparedQuery = prepareQuery(rest, component)
    //console.log('prepared', preparedQuery)
    const results = executeQuery(preparedQuery, useWorld)
    //console.log('results', results)
    const returns = resolveReturns(results)
    //console.log('returns', returns)
    return returns
  })
})

